# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

from collections import namedtuple
from string import ascii_letters, digits

from hypothesis import strategies as st

Strategy = namedtuple("Strategy", ["string", "value"])


@st.composite
def floats(draw):
    """Generate floating point numbers in various formats.

    A composite testing strategy to generate floats in the formats
    understood by the `float_t` atom.
    """
    number = draw(st.floats(allow_nan=False, allow_infinity=False))
    fmt = draw(
        st.sampled_from(
            ["{:.20f}", "{:.20e}", "{:.20E}", "{:+.20f}", "{:+.20e}", "{:+.20E}"]
        )
    )
    return Strategy(fmt.format(number), number)


@st.composite
def complex_numbers(draw):
    """Generate complex numbers in various formats.

    A composite testing strategy to generate complex numbers in the formats
    understood by the `complex_t` atom.
    """
    number = draw(st.complex_numbers(allow_nan=False, allow_infinity=False))
    has_real = draw(st.booleans())
    w_space = draw(st.booleans())
    real_fmt = draw(st.sampled_from(["{:.20g}", "{:.20G}"]))
    imag_fmt = draw(
        st.sampled_from(["{:+.20g}*j", "{:+.20G}*j", "{:+.20g}*J", "{:+.20G}*J"])
    )
    number_as_string = ""
    if has_real:
        if w_space:
            fmt = real_fmt + " " + imag_fmt
        else:
            fmt = real_fmt + imag_fmt
        number_as_string = fmt.format(number.real, number.imag)
    else:
        number = complex(0.0, number.imag)
        fmt = imag_fmt
        number_as_string = fmt.format(number.imag)
    return Strategy(number_as_string, number)


@st.composite
def unquoted_str(draw):
    """Generate strings compatible with our definition of an unquoted string."""

    start = draw(st.text(alphabet=(ascii_letters + "_"), min_size=1))
    body = draw(st.text(alphabet=(ascii_letters + digits + "_")))

    return start + body


@st.composite
def list_of_floats(draw, *, start: str = "[", end: str = "]", delimiter: str = ","):
    """Generate list of floating point numbers in various formats.

    A composite testing strategy to generate lists of floats in the formats
    understood by the list atoms that can be generated by the `make_list_t`
    function.
    """
    lst = draw(st.lists(floats(), min_size=1))

    list_as_string = start + delimiter.join((x.string for x in lst)) + end
    numbers = [x.value for x in lst]

    return Strategy(list_as_string, numbers)


@st.composite
def list_of_complex_numbers(
    draw, *, start: str = "[", end: str = "]", delimiter: str = ","
):
    """Generate list of complex numbers in various formats.

    A composite testing strategy to generate lists of complex numbers in the formats
    understood by the list atoms that can be generated by the `make_list_t`
    function.
    """
    lst = draw(st.lists(complex_numbers(), min_size=1))

    list_as_string = start + delimiter.join((x.string for x in lst)) + end
    numbers = [x.value for x in lst]

    return Strategy(list_as_string, numbers)


@st.composite
def list_of_unquoted_str(
    draw, *, start: str = "[", end: str = "]", delimiter: str = ","
):
    """Generate list of unquoted strings."""
    lst = draw(st.lists(unquoted_str(), min_size=1))

    list_as_string = start + delimiter.join((x for x in lst)) + end
    strings = [x for x in lst]

    return Strategy(list_as_string, strings)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

import json
import math
from io import StringIO

import pytest

from parselglossy.grammars import getkw, lexer

# fmt: off
reference = {
    'int': 42
  , 'dbl': math.pi
  , 'bool': True
  , 'str': 'fooffa'
  , 'int_array': [42]
  , 'dbl_array': [math.pi, math.e, 2.0*math.pi]
  , 'bool_array': [True, True, True, False, True, False]
  , 'str_array': ["foo", "bar", "lorem", "IpSuM"]
  , 'raw': "H 0.0 0.0 0.0\nF 1.0 1.0 1.0\n"
}
# fmt: on


def contents():
    contents = """/* This is a comment */
int = 42
// This is another comment
dbl = {PI}
bool = on
str = "fooffa"

! Yet another comment
$raw
H 0.0 0.0 0.0
F 1.0 1.0 1.0
$end

# I love comments!
int_array = [42]
dbl_array = [{PI}, {E}, {TAU}]
bool_array = [on, true, yes, False, True, false]
str_array = [foo, bar, "lorem", "IpSuM"]
"""
    return contents


@pytest.fixture
def keywords():
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    keys = """{CONTENTS}
"""
    return keys.format(CONTENTS=stuff)


def test_keyword(keywords):
    """Test an input made only of keywords."""
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, keywords)

    assert tokens == reference
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == reference


def section(name):
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    sect = """{NAME} {{
    {CONTENTS}
}}
"""
    return sect.format(NAME=name, CONTENTS=stuff)


@pytest.mark.parametrize("name", ["topsect", "foo<bar>"])
def test_section(name):
    """Test an input made of one section, tagged or untagged."""
    ref_dict = {name: dict(reference)}
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, section(name))

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


@pytest.fixture
def flat_sections():
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    sects = """topsect {{
    {CONTENTS}
}}

foo<bar> {{
    {CONTENTS}
}}
"""
    return sects.format(CONTENTS=stuff)


def test_flat_sections(flat_sections):
    """Test an input made of two unnested sections, tagged or untagged."""
    ref_dict = {"topsect": dict(reference), "foo<bar>": dict(reference)}
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, flat_sections)

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


@pytest.fixture
def nested_sections():
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    sects = """topsect {{
    {CONTENTS}

    foo<bar> {{
        {CONTENTS}
    }}
}}
"""
    return sects.format(CONTENTS=stuff)


def test_nested_sections(nested_sections):
    """Test an input made of two nested sections, tagged or untagged."""
    ref_dict = {"topsect": dict(reference)}
    ref_dict["topsect"]["foo<bar>"] = dict(reference)
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, nested_sections)

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


def keywords_and_section(name):
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    template = """/* This is a comment */
int = 42
// This is another comment
dbl = {PI}
bool = on
str = "fooffa"

{NAME} {{
    {CONTENTS}
}}

! Yet another comment
$raw
H 0.0 0.0 0.0
F 1.0 1.0 1.0
$end

# I love comments!
int_array = [42]
dbl_array = [{PI}, {E}, {TAU}]
bool_array = [on, true, yes, False, True, false]
str_array = [foo, bar, "lorem", "IpSuM"]
"""
    inp = template.format(
        PI=math.pi, E=math.e, TAU=2.0 * math.pi, NAME=name, CONTENTS=stuff
    )
    return inp


@pytest.mark.parametrize("name", ["topsect", "foo<bar>"])
def test_keywords_and_section(name):
    """Test an input made of keywords, one section, tagged or untagged, and more keywords."""
    ref_dict = dict(reference)
    ref_dict[name] = dict(reference)
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, keywords_and_section(name))

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


@pytest.fixture
def keywords_and_flat_sections():
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    template = """/* This is a comment */
int = 42
// This is another comment
dbl = {PI}
bool = on
str = "fooffa"

topsect {{
    {CONTENTS}
}}

! Yet another comment
$raw
H 0.0 0.0 0.0
F 1.0 1.0 1.0
$end

foo<bar> {{
    {CONTENTS}
}}

# I love comments!
int_array = [42]
dbl_array = [{PI}, {E}, {TAU}]
bool_array = [on, true, yes, False, True, false]
str_array = [foo, bar, "lorem", "IpSuM"]
"""
    inp = template.format(PI=math.pi, E=math.e, TAU=2.0 * math.pi, CONTENTS=stuff)
    return inp


def test_keywords_and_flat_sections(keywords_and_flat_sections):
    """Test an input made of keywords and two unnested sections, interspersed."""
    ref_dict = dict(reference)
    ref_dict["topsect"] = dict(reference)
    ref_dict["foo<bar>"] = dict(reference)
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, keywords_and_flat_sections)

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


@pytest.fixture
def keywords_and_nested_sections():
    stuff = contents().format(PI=math.pi, E=math.e, TAU=2.0 * math.pi)
    template = """/* This is a comment */
int = 42
// This is another comment
dbl = {PI}
bool = on
str = "fooffa"

! Yet another comment
$raw
H 0.0 0.0 0.0
F 1.0 1.0 1.0
$end

topsect {{
    {CONTENTS}

    foo<bar> {{
        {CONTENTS}
    }}
}}

# I love comments!
int_array = [42]
dbl_array = [{PI}, {E}, {TAU}]
bool_array = [on, true, yes, False, True, false]
str_array = [foo, bar, "lorem", "IpSuM"]
"""
    inp = template.format(PI=math.pi, E=math.e, TAU=2.0 * math.pi, CONTENTS=stuff)
    return inp


def test_keywords_and_nested_sections(keywords_and_nested_sections):
    """Test an input made of keywords and two nested sections, interspersed."""
    ref_dict = dict(reference)
    ref_dict["topsect"] = dict(reference)
    ref_dict["topsect"]["foo<bar>"] = dict(reference)
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, keywords_and_nested_sections)

    assert tokens == ref_dict
    # dump to JSON
    getkw_json = StringIO()
    json.dump(tokens, getkw_json, indent=4)
    del tokens

    # load from JSON
    tokens = json.loads(getkw_json.getvalue())

    assert tokens == ref_dict


@pytest.fixture
def data_only_section():
    stuff = """molecule {
$coords
H  0.0000  0.0000 -0.7000
H  0.0000  0.0000  0.7000
$end
}
    """
    return stuff


def test_data_only_section(data_only_section):
    ref = {
        "molecule": {"coords": "H  0.0000  0.0000 -0.7000\nH  0.0000  0.0000  0.7000\n"}
    }
    grammar = getkw.grammar()
    tokens = lexer.parse_string_to_dict(grammar, data_only_section)

    assert tokens == ref



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Console script for parselglossy."""

from pathlib import Path
from typing import List, Union

import click

from . import __version__
from .api import generate


@click.group()
@click.version_option(prog_name="parselglossy", version=__version__)
def cli(args=None):
    """Console script for parselglossy."""
    return 0


@click.command(name="generate")
@click.option(
    "--template",
    type=click.Path(exists=False, dir_okay=False, resolve_path=True),
    metavar="<template>",
    help="which validation template to use",
)
@click.option(
    "--target",
    type=click.Path(exists=False, dir_okay=True, resolve_path=True),
    default="input_parser",
    show_default=True,
    metavar="<target>",
    help="name of the generated Python module",
)
@click.option(
    "--grammar",
    "-g",
    default=["standard"],
    multiple=True,
    show_default=True,
    metavar="<grammar>",
    help="which grammar to use, specify multiple times for multiple external files",
)
@click.option(
    "--tokenize",
    type=str,
    default=None,
    metavar="<tokenize>",
    help="tokenizing commands for the custom grammar",
)
@click.option(
    "--docfile",
    type=str,
    default="input.rst",
    show_default=True,
    metavar="<header>",
    help="name for the generated documentation file",
)
@click.option(
    "--doc-header",
    type=str,
    default="Input parameters",
    show_default=True,
    metavar="<header>",
    help="header for the documentation page",
)
def _generate(
    template: str,
    target: str,
    grammar: Union[str, List[str]],
    tokenize: str,
    docfile: str,
    doc_header: str,
):
    """Generate parser Python module and documentation from a grammar and a
    validation template.

    \b
    Parameters
    ----------
    template : str
        Which validation template to use.
    target : str
        Name of the generated Python moduel.
        Defaults to ``input_parameters`` in the current working directory.
    grammar : str
        Which grammar to use, can be a list of files.
        Defaults to ``standard``.
    tokenize : str
        Tokenizing commands to run with the custom grammar.
        Defaults to ``None``.
    docfile : str
        The name of the documentation file for the input.
        Defaults to `input.rst`.
    doc_header : str
        Header for the documentation page.
        Defaults to ``Input parameters``.
    """

    if not docfile:
        docfile = "input.rst"

    if "standard" in grammar or "getkw" in grammar:
        grammar_ = grammar[0]
    else:
        grammar_ = [Path(_).resolve() for _ in grammar]  # type: ignore

    generate(
        template=template,
        where=target,
        grammar=grammar_,
        tokenize=tokenize,
        docfile=docfile,
        doc_header=doc_header,
    )


cli.add_command(_generate)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Constants used by the generator."""

from datetime import date
from pathlib import Path
from re import DOTALL, search
import pprint

INIT_PY = f"""# -*- coding: utf-8 -*-

# This file was automatically generated by parselglossy on {date.today()}
# Editing is *STRONGLY DISCOURAGED*
"""
"""str: Content of generated __init__.py"""

README = f"""This file was automatically generated by parselglossy on {date.today()}
Editing is *STRONGLY DISCOURAGED*
"""
"""str: Content of generated README.md"""


CLI_PY = f"""# -*- coding: utf-8 -*-

# This file was automatically generated by parselglossy on {date.today()}
# Editing is *STRONGLY DISCOURAGED*

import argparse

from .api import parse
from .plumbing.utils import default_outfile


def cli():
    cli = argparse.ArgumentParser()
    cli.add_argument("infile", help="the input file to parse")
    cli.add_argument(
        "--dump-ir",
        dest="dumpir",
        help="whether to dump intermediate representation to JSON",
        action="store_true",
        default=False,
    )
    cli.add_argument(
        "--outfile",
        dest="outfile",
        help="name or path for the parsed JSON output file",
        type=str,
        action="store",
    )

    args = cli.parse_args()

    if args.outfile is None:
        outfile = default_outfile(fname=args.infile, suffix="_fr.json")

    parse(infile=args.infile, dump_ir=args.dumpir, outfile=outfile)
"""
"""str: Content of generated cli.py"""


def get_parse_string_to_dict() -> str:
    with (Path(__file__).parent.absolute() / "grammars/lexer.py").open("r") as buf:
        m = search(
            r"\# ->->-> SNIP <-<-<-\n(.*)\n\# -<-<-< SNAP >->->-", buf.read(), DOTALL,
        )
    if m is not None:
        return f"""\n
from pathlib import Path
from typing import Union

from .exceptions import ParselglossyError
from .utils import JSONDict, flatten_list, dict_to_list


{m.group(1)}"""
    else:
        raise RuntimeError("Nothing extracted from lexer.py!")


def lexer_py(lexer_str: str) -> str:
    return f"""# -*- coding: utf-8 -*-

# This file was automatically generated by parselglossy on {date.today()}
# Editing is *STRONGLY DISCOURAGED*

import json
from pathlib import Path
from typing import Optional, Union

from .utils import ComplexEncoder, JSONDict, path_resolver


def lex_from_str(*,
                 in_str: Union[str, Path],
                 ir_file: Optional[Union[str, Path]] = None
) -> JSONDict:
    \"\"\"Run grammar on input string.

    Parameters
    ----------
    in_str : Union[str, Path]
         The string to be parsed.
    ir_file : Optional[Union[str, Path]]
         File to write intermediate representation to (JSON format).
         None by default, which means file is not written out.

    Returns
    -------
    The contents of the input string as a dictionary.
    \"\"\"

    {lexer_str}

    if ir_file is not None:
        ir_file = path_resolver(ir_file)
        with ir_file.open("w") as out:
            json.dump(ir, out, cls=ComplexEncoder, indent=4)

    return ir
"""


def api_py(stencil):
    return f"""# -*- coding: utf-8 -*-

# This file was automatically generated by parselglossy on {date.today()}
# Editing is *STRONGLY DISCOURAGED*

import json
from pathlib import Path
from typing import Optional, Union

from .plumbing import lexer
from .plumbing.utils import JSONDict, as_complex, path_resolver, copier
from .plumbing.validation import validate_from_dicts


def lex(
    infile: Union[str, Path],
    ir_file: Optional[Union[str, Path]] = None,
) -> JSONDict:
    \"\"\"Run grammar of choice on input string.

    Parameters
    ----------
    infile : Union[str, Path]
        The string to be parsed.
    ir_file : Optional[Union[str, Path]]
        File to write intermediate representation to (JSON format).
        None by default, which means file is not written out.

    Returns
    -------
    The contents of the input string as a dictionary.
    \"\"\"

    infile = path_resolver(infile)
    if ir_file is not None:
        ir_file = path_resolver(ir_file)

    with infile.open("r") as f:
        ir = lexer.lex_from_str(in_str=f.read(), ir_file=ir_file)

    return ir


def validate(
    infile: Union[str, Path],
    fr_file: Optional[Union[str, Path]] = None,
) -> JSONDict:
    \"\"\"Validate intermediate representation into final representation.

    Parameters
    ----------
    infile : Union[str, Path]
        The file with the intermediate representation (JSON format).
    fr_file : Optional[Union[str, Path]]
        File to write final representation to (JSON format).
        None by default, which means file is not written out.

    Returns
    -------
    The validated input as a dictionary.
    \"\"\"

    infile = path_resolver(infile)
    with infile.open("r") as f:
        ir = json.load(f, object_hook=as_complex)

    if fr_file is not None:
        fr_file = path_resolver(fr_file)

    return validate_from_dicts(ir=ir, template=stencil(), fr_file=fr_file)


def parse(
    infile: Union[str, Path],
    outfile: Optional[Union[str, Path]] = None,
    dump_ir: bool = False,
) -> JSONDict:
    \"\"\"Parse input file.

    Parameters
    ----------
    infile : Union[str, Path]
        The input file to be parsed.
    outfile : Optional[Union[str, Path]]
        The output file.
        Defaults to ``None``, which means writing to ``<infile>_fr.json``.
    dump_ir : bool
        Whether to write out the intermediate representation to file (JSON format).
        False by default. If true the filename if ``<infile>_ir.json``

    Returns
    -------
    The validated input as a dictionary.
    \"\"\"

    infile = path_resolver(infile)

    ir_file = None  # type: Optional[Path]
    if dump_ir:
        ir_file = path_resolver(Path(infile).stem + "_ir.json")

    ir = lex(infile=infile, ir_file=ir_file)

    if outfile is not None:
        outfile = path_resolver(outfile)

    return validate_from_dicts(ir=ir, template=stencil(), fr_file=outfile)


def stencil() -> JSONDict:
    return {pprint.pformat(stencil, indent=4)}
"""



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Top-level functions for parselglossy."""

import json
from pathlib import Path
from typing import Optional, Union

from ..exceptions import ParselglossyError
from ..utils import ComplexEncoder, JSONDict, dict_to_list, flatten_list, path_resolver
from . import getkw


# ->->-> SNIP <-<-<-
def parse_string_to_dict(lexer, s: Union[str, Path]) -> JSONDict:
    """Helper function around parseString(s).asDict()
    that checks whether some keywords or sections were accidentally
    repeated and shadowing earlier keywords/sections.

    Parameters
    ----------
    lexer : JSONDict
         Nested dictionary
    s : Union[str, Path]
         String to parse

    Returns
    -------
    tokes_dict: JSONDict
         Dictionary of tokens

    Raises
    ------
    :exc:`ParselglossyError`
    """
    tokens_dict = lexer.parseString(s).asDict()
    tokens_list = lexer.parseString(s).asList()
    if flatten_list(tokens_list) != flatten_list(dict_to_list(tokens_dict)):
        raise ParselglossyError("A keyword is repeated. Please check your input.")
    return tokens_dict


# -<-<-< SNAP >->->-


def lex_from_str(
    *,
    in_str: Union[str, Path],
    grammar: str = "standard",
    ir_file: Optional[Union[str, Path]] = None,
) -> JSONDict:
    """Run grammar of choice on input string.

    Parameters
    ----------
    in_str : Union[str, Path]
         The string to be parsed.
    grammar : str
         Grammar to be used. Defaults to "standard".
    ir_file : Optional[Union[str, Path]]
         File to write intermediate representation to (JSON format).
         None by default, which means file is not written out.

    Returns
    -------
    The contents of the input string as a dictionary.

    Raises
    ------
    :exc:`ParselglossyError`
    """

    try:
        lexer = dispatch_grammar(grammar)
    except KeyError:
        raise ParselglossyError(f"Grammar {grammar} not available.")

    ir = parse_string_to_dict(lexer, in_str)

    if ir_file is not None:
        ir_file = path_resolver(ir_file)
        with ir_file.open("w") as out:
            json.dump(ir, out, cls=ComplexEncoder, indent=4)

    return ir


def dispatch_grammar(grammar: str):
    available_grammars = {
        "getkw": getkw.grammar(),
        "standard": getkw.grammar(has_complex=True),
    }
    return available_grammars[grammar]
